//================================================================================
// GameMutator.
//
// Copyright(c) 2003, Winged Unicorn.
//================================================================================
class GameMutator extends MCoopBaseMutator
	abstract
	config(MCoop2);


struct RespOpts
{
	var() config bool bRespawnAmmo;
	var() config bool bRespawnHealth;
	var() config bool bRespawnInventory;
	var() config bool bRespawnWeapon;
	var() config float RespawnTimeAmmo;
	var() config float RespawnTimeHealth;
	var() config float RespawnTimeInventory;
	var() config float RespawnTimeWeapon;
};

struct PlayerOpts
{
	var() config bool bDenyNameChange;
	var() config bool bSetDefaultInventory;
	var() config bool bSetDefaultHealth;
	var() config bool bSetDefaultInvCharge;
	var() config bool bSetDefaultAmmo;
};

struct ADV_Network
{
	var() config private bool bOptimizeMaxPlayers;
	var() config private bool bOptimizeConnection;
	var() config private int ServerUploadSpeed;
	var() config private int DefaultMinClientRate;
	var() config private int DefaultMaxClientRate;
};

var(AdvancedNetwork) config private ADV_Network AdvancedNetwork;
var(Security) globalconfig private string AdminPassword;
var(Security) globalconfig private string AdminChildPassword;
var(Security) globalconfig private string AdminBabyPassword;
var(Security) globalconfig private string GamePassword;
var(Security) globalconfig private int MaxFailedAdminLogins;
var(Security) globalconfig bool bSafeAdmins;
var(Security) globalconfig string AdminName;
var(PlayerControl) config PlayerOpts PlayerOptions;
var(GameControl) globalconfig bool bSpecialEffects;
var(GameControl) config bool bAlwaysPickUp;
var(GameControl) config bool bAlwaysReconnect;
var(GameControl) config RespOpts RespawnOptions;
var(GameControl) config array<string> Scripts;

var(MapControl) config bool bCustomMapList;
var(MapControl) class<MCoopMapList> MapListType;
var bool bEnableMapTimer;

var PawnsInfo PawnsInfo;
var IPLogger IPLogger;
var ChatLogger ChatLogger;

var localized string BanError;
var localized string TempBanError;
var private int FailedAdminLogins;

// run-time vars used by MCoopDLL
var private int	IPPreLoginOffset,
		IPLoginOffset,
		OptPreLoginOffset;

// version info :)
const Version = "2.0";


function ServerLog(string Msg)
{
	Log("** "$ Msg $" **", 'MCoop');
}

function WarnLog(string Msg)
{
	Log(">> WARNING!! "$ Msg $"! <<", 'MCoop');
}

function AdminMsg(coerce string Msg)
{
	local Pawn aPawn;

	for ( aPawn=Level.PawnList; aPawn!=None; aPawn=aPawn.nextPawn )
		if ( aPawn.bIsPlayer && (MCoopUnrealIPlayer(aPawn) != None) && (PlayerPawn(aPawn).bAdmin || MCoopUnrealIPlayer(aPawn).bAdminChild) )
			PlayerPawn(aPawn).ClientMessage("** MCoop: "@ Msg @" **", 'Pickup', True);
}

function AdminLogin(MCoopUnrealIPlayer Requester, string Password, bool bNotify)
{
	if ( FailedAdminLogins >= MaxFailedAdminLogins )
	{
		Requester.ClientMessage("Maximum limit of failed AdminLogins reached. Aborting...");
		return;
	}

	if ( (AdminPassword != "") && (Password == AdminPassword) && !Requester.bAdmin )
	{
		Requester.bAdmin      = True;
		Requester.bAdminChild = False;
		Requester.bAdminBaby  = False;
		if ( bNotify )
			BroadcastMessage(Requester.PlayerReplicationInfo.PlayerName $" is now Administrator", True, 'CriticalEvent');
		Requester.ClientMessage("You are now Admin :)");
		ServerLog(Requester.PlayerReplicationInfo.PlayerName $" logged in as Admin");
	}
	else if ( (AdminChildPassword != "") && (Password == AdminChildPassword) && !Requester.bAdminChild )
	{
		Requester.bAdmin      = False;
		Requester.bAdminChild = True;
		Requester.bAdminBaby  = False;
		if ( bNotify )
			BroadcastMessage(Requester.PlayerReplicationInfo.PlayerName $" is now AdminChild", True, 'CriticalEvent');
		Requester.ClientMessage("You are now AdminChild :)");
		ServerLog(Requester.PlayerReplicationInfo.PlayerName $" logged in as AdminChild");
	}
	else if ( (AdminBabyPassword != "") && (Password == AdminBabyPassword) && !Requester.bAdminBaby )
	{
		Requester.bAdmin      = False;
		Requester.bAdminChild = False;
		Requester.bAdminBaby  = True;
		if ( bNotify )
			BroadcastMessage(Requester.PlayerReplicationInfo.PlayerName $" is now AdminBaby", True, 'CriticalEvent');
		Requester.ClientMessage("You are now AdminBaby :)");
		ServerLog(Requester.PlayerReplicationInfo.PlayerName $" logged in as AdminBaby");
	}
	else
		FailedAdminLogins++;
}

event Destroyed()
{
	if ( IPLogger != None )
	{
		IPLogger.Destroy();
		IPLogger = None;
	}
	if ( ChatLogger != None )
	{
		ChatLogger.Destroy();
		ChatLogger = None;
	}
	if ( PawnsInfo != None )
	{
		PawnsInfo.Destroy();
		PawnsInfo = None;
	}

	Super.Destroyed();
}

//================================================================================
//------------------------------ Begin Play --------------------------------------

// dynamically set the DefaultProperties of the current game type
function ServerSetDefaultProperties()
{
}

function ServerPreBeginPlay()
{
	if ( bSpecialEffects && (Level.NetMode == NM_DedicatedServer) )
		Level.NetMode = NM_ListenServer;
}

function ServerPostBeginPlay()
{
	local class<PawnsList> PawnsListClass;
	local PawnsList PawnsList;
	local string NextPlayerClass;
	local int i;
	local MCoopMapList myList;

	Log("***********************************************************************************************", 'MCoop');
	Log(" MonsterCoop mod now running! ", 'MCoop');
	Log(" by Winged Unicorn - version "$ Version, 'MCoop');
	Log(" Thanks go to: ", 'MCoop');
	Log(" - UsAaR33, -=O-HbG-O=-, }TCP{Wolf and Boodaflow ;)", 'MCoop');
	Log(" A special thank to: ", 'MCoop');
	Log(" - Tammy, aka sniperwolf, and Zora :)", 'MCoop');
	Log("***********************************************************************************************", 'MCoop');

	// load all player classes
	PawnsListClass = class<PawnsList>(DynamicLoadObject(MCoopGame.PawnsListClass, class'Class'));
	if ( PawnsListClass == None )
	{
		WarnLog("GameMutator.ServerPostBeginPlay couldn't load class '"$ MCoopGame.PawnsListClass $"'");
		PawnsListClass = class'MCoop2.PawnsList';
		MCoopGame.PawnsListClass = "MCoop2.PawnsList";
	}

	PawnsList = Spawn(PawnsListClass);
	if ( PawnsList != None )
	{
		NextPlayerClass = PawnsList.GetPlayerClassInfo(i);
		while ((NextPlayerClass != "") && (i < 200))
		{
			DynamicLoadObject(NextPlayerClass, class'Class');
			NextPlayerClass = PawnsList.GetPlayerClassInfo(++i);
		}

		PawnsList.Destroy();
	}

	// handle network options
	if ((Level.NetMode == NM_DedicatedServer) || (Level.NetMode == NM_ListenServer))
	{
		// set fast time-out
		ConsoleCommand("set ipdrv.tcpnetdriver InitialConnectTimeout 30");
		ConsoleCommand("set ipdrv.tcpnetdriver ConnectionTimeout 15");

		// optimize connection
		if ( AdvancedNetwork.bOptimizeConnection )
		{
			// MaxClientRate (will be changed dinamically)
			if ( (AdvancedNetwork.DefaultMaxClientRate != 0) && (AdvancedNetwork.DefaultMaxClientRate != int(ConsoleCommand("get ipdrv.tcpnetdriver MaxClientRate"))) )
			{
				ConsoleCommand("set ipdrv.tcpnetdriver MaxClientRate "$ AdvancedNetwork.DefaultMaxClientRate);
				Log("-- New MaxClientRate... "$ AdvancedNetwork.DefaultMaxClientRate $" --", 'MCoop');
			}

			// MaxTicksPerSecond / NetServerMaxTickRate (should always be 20)
			if ( Level.EngineVersion != "226" )
			{
				if ( int(ConsoleCommand("get ipdrv.tcpnetdriver MaxTicksPerSecond")) != 20 )
				{
					ConsoleCommand("set ipdrv.tcpnetdriver MaxTicksPerSecond 20");
					Log("-- MaxTicksPerSecond... 20 --", 'MCoop');
				}
			}
			else
			{
				if ( int(ConsoleCommand("get ipdrv.tcpnetdriver NetServerMaxTickRate")) != 20 )
				{
					ConsoleCommand("set ipdrv.tcpnetdriver NetServerMaxTickRate 20");
					Log("-- NetServerMaxTickRate... 20 --", 'MCoop');
				}
			}
		}
	}

	// manage server's name
	if ( InStr(Caps(MCoopGame.GameReplicationInfo.ServerName), "MCOOP") == -1 )
		MCoopGame.GameReplicationInfo.ServerName = MCoopGame.GameReplicationInfo.ServerName @" [MCoop2]";
	if ( (GamePassword != "") && (InStr(Caps(MCoopGame.GameReplicationInfo.ServerName), "PWD") == -1) )
		MCoopGame.GameReplicationInfo.ServerName = MCoopGame.GameReplicationInfo.ServerName $"[Pwd]";

	// spawn ServerTools actors
	IPLogger = Spawn(class'IPLogger');
	ChatLogger = Spawn(class'ChatLogger');
	if ( (ChatLogger != None) && ChatLogger.bLogChat && (ChatLogger.MCoopLogFile == None) )
	{
		ChatLogger.MCoopLogFile = Spawn(class'MCoopLogFile');
		if ( ChatLogger.MCoopLogFile != None )
			ChatLogger.MCoopLogFile.StartLog();
	}

	// initialize Teams
	for (i=0; i<2; i++)
	{
		MCoopGame.Teams[i] = Spawn(class'TeamInfo');
		MCoopGame.Teams[i].Size  = 0;
		MCoopGame.Teams[i].Score = 0;
		MCoopGame.Teams[i].TeamName  = MCoopGame.TeamColor[i];
		MCoopGame.Teams[i].TeamIndex = i;
	}

	// set TimeLimit
	if ((MapListType != None) && bCustomMapList )
	{
		myList = Spawn(MapListType);
		MCoopGame.TimeLimit = myList.GetTimeLimit();
		if ( MCoopGame.TimeLimit > 0 )
		{
			bEnableMapTimer = True;
			MCoopGame.RemainingTime = 60 * MCoopGame.TimeLimit;
			ServerLog("Map Timer ENABLED! Time limit is... "$ MCoopGame.TimeLimit $" minutes");
		}

		myList.Destroy();
	}

	MCoopGame.CurrentID++;

	// new feature asked by Hyper: it can load a bunch of scripts at startup.
	// admins can define as many scripts as they want, since they're stored in a dynamic array
	for ( i=0; i < class'MCoopDLL'.static.SizeOf(Scripts); i++ )
		Log(class'MCoopDLL'.static.Get(Scripts, i) $"... "$ ConsoleCommand(class'MCoopDLL'.static.Get(Scripts, i)), 'MCoop');
}

function ServerInitGameReplicationInfo()
{
	if ((MCoopGame.GameReplicationInfo != None) && (MCoopGameReplicationInfo(MCoopGame.GameReplicationInfo) != None))
		MCoopGameReplicationInfo(MCoopGame.GameReplicationInfo).bDeathMatchGame = MCoopGame.bDeathMatch;

	// spawn Pawns Info
	InitPawnsInfo();
}

function InitPawnsInfo()
{
	local NavigationPoint N;
	local int i;

	PawnsInfo = Spawn(class'PawnsInfo');

	for ( N=Level.NavigationPointList; N!=None; N=N.nextNavigationPoint )
	{
		if ( N.IsA('PathNode') )
		{
			for ( i=0; i<24; i++ )
				PawnsInfo.Pawns[i] = 1;

			break;
		}
	}
}

//
// Initialize the game.
// this is called before actors' PreBeginPlay.
//
function ServerInitGame(string Options, out string Error)
{
	//*******************************************
	// do NOT change these values if not needed!!
	// moreover, remember to add a Super call if
	// you plan to modify this function...
	//*******************************************
	if ( Level.EngineVersion == "224" )
	{
		IPPreLoginOffset  = 7358;
		OptPreLoginOffset = 3756;

		if ((Team(Self) != None) && (MonsterCoop(Self) == None))
			IPLoginOffset = 15922;
		else
			IPLoginOffset = 13554;
	}
	else if ( Level.EngineVersion == "225" )
	{
		IPPreLoginOffset  = 7346;
		OptPreLoginOffset = 3764;

		if ((Team(Self) != None) && (MonsterCoop(Self) == None))
			IPLoginOffset = 15910;
		else
			IPLoginOffset = 13542;
	}
	else
	{
		IPPreLoginOffset  = 7326;
		OptPreLoginOffset = 3764;

		if ((Team(Self) != None) && (MonsterCoop(Self) == None))
			IPLoginOffset = 15890;
		else
			IPLoginOffset = 13522;
	}
}

// called by GameInfo.IsRelevant
// NOTE: this function does NOT return a value (thus can't destroy actors)
// so use a standard mutator if you need to...
function ServerIsRelevant(Actor Other)
{
	if ( Bots(Other) != None )
		Bots(Other).PlayerReplicationInfoClass = class'MCoop2.MCoopPlayerReplicationInfo';
}

function ServerSetGameSpeed(float T)
{
	MCoopGame.GameSpeed = FMax(T, 0.1);
	Level.TimeDilation = MCoopGame.GameSpeed;
}

function ServerTimer()
{
	if ( bEnableMapTimer )
	{
		if ( MCoopGame.bGameEnded )
		{
			MCoopGame.RemainingTime--;
			if ( MCoopGame.RemainingTime < -7 )
				MCoopGame.RestartGame();
		}
		else if ( MCoopGame.TimeLimit > 0 )
		{
			MCoopGame.RemainingTime--;
			switch (MCoopGame.RemainingTime)
			{
				case 300:
					MCoopGame.BroadcastMessage(MCoopGame.TimeMessage[0], True, 'CriticalEvent');
					Log(MCoopGame.TimeMessage[0]);
					break;
				case 240:
					MCoopGame.BroadcastMessage(MCoopGame.TimeMessage[1], True, 'CriticalEvent');
					Log(MCoopGame.TimeMessage[1]);
					break;
				case 180:
					MCoopGame.BroadcastMessage(MCoopGame.TimeMessage[2], True, 'CriticalEvent');
					Log(MCoopGame.TimeMessage[2]);
					break;
				case 120:
					MCoopGame.BroadcastMessage(MCoopGame.TimeMessage[3], True, 'CriticalEvent');
					Log(MCoopGame.TimeMessage[3]);
					break;
				case 60:
					MCoopGame.BroadcastMessage(MCoopGame.TimeMessage[4], True, 'CriticalEvent');
					Log(MCoopGame.TimeMessage[4]);
					break;
				case 30:
					MCoopGame.BroadcastMessage(MCoopGame.TimeMessage[5], True, 'CriticalEvent');
					Log(MCoopGame.TimeMessage[5]);
					break;
				case 10:
					MCoopGame.BroadcastMessage(MCoopGame.TimeMessage[6], True, 'CriticalEvent');
					Log(MCoopGame.TimeMessage[6]);
					break;
				case 5:
					MCoopGame.BroadcastMessage(MCoopGame.TimeMessage[7], True, 'CriticalEvent');
					Log(MCoopGame.TimeMessage[7]);
					break;
				case 4:
					MCoopGame.BroadcastMessage(MCoopGame.TimeMessage[8], True, 'CriticalEvent');
					Log(MCoopGame.TimeMessage[8]);
					break;
				case 3:
					MCoopGame.BroadcastMessage(MCoopGame.TimeMessage[9], True, 'CriticalEvent');
					Log(MCoopGame.TimeMessage[9]);
					break;
				case 2:
					MCoopGame.BroadcastMessage(MCoopGame.TimeMessage[10], True, 'CriticalEvent');
					Log(MCoopGame.TimeMessage[10]);
					break;
				case 1:
					MCoopGame.BroadcastMessage(MCoopGame.TimeMessage[11], True, 'CriticalEvent');
					Log(MCoopGame.TimeMessage[11]);
					break;
				case 0:
					MCoopGame.BroadcastMessage(MCoopGame.TimeMessage[12], True, 'CriticalEvent');
					Log(MCoopGame.TimeMessage[12]);
					break;
			}

			if ( MCoopGame.RemainingTime <= 0 )
				MCoopGame.EndGame("timelimit");
		}
	}
}

//------------------------------ Login Events --------------------------------------
//
// Accept or reject a player on the server.
// Fails login if you set the Error to a non-empty string.
//
function ServerPreLogin(string Options, out string Error)
{
	local string InName, InClass, InSkin, InPassword, PlayerIP, Port;
	local bool bPendingAdmin, bLoginProtectorFound, bIsIn;
	local byte InTeam, PawnType;
	local LoginProtector LoginProtector;
	local Pawn aPawn;

	Error = "";

	// get Options
	class'MCoopDLL'.static.ParseOptions( OptPreLoginOffset, InName, InClass, InSkin, InTeam, InPassword );

	// check for passwords
	bPendingAdmin = (InPassword == AdminPassword);
	if ( (GamePassword != "") && (InPassword != GamePassword) && ((AdminPassword == "") || (InPassword != AdminPassword)) )
	{
		if (InPassword == "")
			Error = MCoopGame.NeedPassword;
		else
			Error = MCoopGame.WrongPassword;
	}

	// get IP
	class'MCoopDLL'.static.GetIP( IPPreLoginOffset, PlayerIP, Port );
	Log("-- PreLogin attempt by player "$ InName $" -- IP Address: "$ PlayerIP $":"$ Port, 'MCoop');

	if ( IPLogger.bEnabled )
	{
		// check for IP ban
		if ( IPLogger.CheckBanIP(InName, PlayerIP) )
		{
			Error = BanError;
			return;
		}

		// check for temp IP ban
		if ( IPLogger.CheckTempBanIP(InName, PlayerIP) )
		{
			Error = TempBanError;
			return;
		}
	}

	// check for fast login
	foreach AllActors(class'LoginProtector', LoginProtector)
	{
		if ( LoginProtector.PlayerIP == PlayerIP )
		{
			bLoginProtectorFound = True;

			if ( !LoginProtector.bAllowLogin )
			{
				if ( LoginProtector.Increment() && IPLogger.bEnabled )
				{
					IPLogger.TempBanIP(PlayerIP, InName);
					Error = TempBanError;
				}
				else
					Error = "You reconnected too quickly, just wait "$ LoginProtector.TimeLeft $" seconds before trying again...";

				return;
			}
			else
			{
				// save the last login options request
				LoginProtector.LoginOptions.InName	= InName;
				LoginProtector.LoginOptions.InClass	= InClass;
				LoginProtector.LoginOptions.SpawnClass	= ModifyLogin(InName, InClass, InTeam, PawnType);
				LoginProtector.LoginOptions.InSkin	= InSkin;
				LoginProtector.LoginOptions.InTeam	= InTeam;
				LoginProtector.LoginOptions.InPassword	= InPassword;
				LoginProtector.LoginOptions.PawnType	= PawnType;

				LoginProtector.GoToState('CountDown');
			}
		}
	}

	if ( !bLoginProtectorFound )
	{
		LoginProtector = Spawn(class'LoginProtector');
		if ( LoginProtector != None )
		{
			// save the last login options request
			LoginProtector.LoginOptions.InName	= InName;
			LoginProtector.LoginOptions.InClass	= InClass;
			LoginProtector.LoginOptions.SpawnClass	= ModifyLogin(InName, InClass, InTeam, PawnType);
			LoginProtector.LoginOptions.InSkin	= InSkin;
			LoginProtector.LoginOptions.InTeam	= InTeam;
			LoginProtector.LoginOptions.InPassword	= InPassword;
			LoginProtector.LoginOptions.PawnType	= PawnType;

			LoginProtector.PlayerIP = PlayerIP;
		}
		else
			WarnLog("GameMutator.ServerPreLogin couldn't spawn class 'MCoopMutator2.LoginProtector'");
	}

	// allow players in the server to reconnect even if server is full
	if ( bAlwaysReconnect )
	{
		for (aPawn=Level.PawnList; aPawn!=None; aPawn=aPawn.nextPawn)
			if ( aPawn.IsA('PlayerPawn') && (aPawn.PlayerReplicationInfo != None)
				&& (MCoopPlayerReplicationInfo(aPawn.PlayerReplicationInfo).PlayerIP == PlayerIP) )
			{
				bIsIn = True;
				break;
			}
	}

	// check if the requested player class is available at the moment
	if ( (PawnType != 255) && (PawnsInfo.Pawns[PawnType] < 1) )
	{
		Error = MCoopGame.FailedSpawnMessage;
		return;
	}

	if ( !bPendingAdmin && (Error == "") )
	{
		if ( (MCoopGame.MaxPlayers > 0) && (MCoopGame.NumPlayers >= MCoopGame.MaxPlayers) && !bIsIn )
		{
			Error = MCoopGame.MaxedOutMessage;
			return;
		}

		if ( MCoopGame.bTeamGame )
		{
			if ( PawnType == 255 )
			{
				if ( MCoopGame.Teams[0].Size >= MCoopGame.MaxTeamSize )
				{
					if ( MCoopGame.Teams[1].Size < MCoopGame.MaxTeamSize )
						Error = "All Human spots used, but you can join as Monster ;)";
					else
						Error = "All Human spots used, try later :(";
					return;
				}
			}
			else
			{
				if ( MCoopGame.Teams[1].Size >= MCoopGame.MaxTeamSize )
				{
					if ( MCoopGame.Teams[0].Size < MCoopGame.MaxTeamSize )
						Error = "All Monster spots used, but you can join as Human ;)";
					else
						Error = "All Monster spots used, try later :(";
					return;
				}
			}
		}
	}

	// tell admins that a new player is about to join
	AdminMsg(InName $" is joining this server...");
}

function class<PlayerPawn> ModifyLogin(string InName, string DesiredClass, out byte InTeam, out byte PawnType)
{
	local class<PlayerPawn> SpawnClass;

	InTeam = 1;

	if ((DesiredClass ~= "MCoop2.MCoopBird") || (InName ~= "bird"))					{ SpawnClass = Class'MCoopBird';		PawnType = 0; }
	else if ((DesiredClass ~= "MCoop2.MCoopBiterFish") || (InName ~= "biterfish"))			{ SpawnClass = Class'MCoopBiterFish';		PawnType = 1; }
	else if ((DesiredClass ~= "MCoop2.MCoopBlob") || (InName ~= "blob"))				{ SpawnClass = Class'MCoopBlob';		PawnType = 2; }
	else if ((DesiredClass ~= "MCoop2.MCoopBrute") || (InName ~= "brute"))				{ SpawnClass = Class'MCoopBrute';		PawnType = 3; }
	else if ((DesiredClass ~= "MCoop2.MCoopBehemoth") || (InName ~= "behemoth"))			{ SpawnClass = Class'MCoopBehemoth';		PawnType = 3; }
	else if ((DesiredClass ~= "MCoop2.MCoopLesserBrute") || (InName ~= "lesserbrute"))		{ SpawnClass = Class'MCoopLesserBrute';		PawnType = 3; }
	else if ((DesiredClass ~= "MCoop2.MCoopDevilFish") || (InName ~= "devilfish"))			{ SpawnClass = Class'MCoopDevilFish';		PawnType = 4; }
	else if ((DesiredClass ~= "MCoop2.MCoopFly") || (InName ~= "fly"))				{ SpawnClass = Class'MCoopFly';			PawnType = 5; }
	else if ((DesiredClass ~= "MCoop2.MCoopGasBag") || (InName ~= "gasbag"))			{ SpawnClass = Class'MCoopGasBag';		PawnType = 6; }
	else if ((DesiredClass ~= "MCoop2.MCoopGiantGasBag") || (InName ~= "giantgasbag"))		{ SpawnClass = Class'MCoopGiantGasBag';		PawnType = 7; }
	else if ((DesiredClass ~= "MCoop2.MCoopGiantManta") || (InName ~= "giantmanta"))		{ SpawnClass = Class'MCoopGiantManta';		PawnType = 8; }
	else if ((DesiredClass ~= "MCoop2.MCoopHorseFly") || (InName ~= "horsefly"))			{ SpawnClass = Class'MCoopHorseFly';		PawnType = 9; }
	else if ((DesiredClass ~= "MCoop2.MCoopKrall") || (InName ~= "krall"))				{ SpawnClass = Class'MCoopKrall';		PawnType = 10; }
	else if ((DesiredClass ~= "MCoop2.MCoopKrallElite") || (InName ~= "krallelite"))		{ SpawnClass = Class'MCoopKrallElite';		PawnType = 10; }
	else if ((DesiredClass ~= "MCoop2.MCoopManta") || (InName ~= "manta"))				{ SpawnClass = Class'MCoopManta';		PawnType = 11; }
	else if ((DesiredClass ~= "MCoop2.MCoopCaveManta") || (InName ~= "cavemanta"))			{ SpawnClass = Class'MCoopCaveManta';		PawnType = 11; }
	else if ((DesiredClass ~= "MCoop2.MCoopMercenary") || (InName ~= "mercenary"))			{ SpawnClass = Class'MCoopMercenary';		PawnType = 12; }
	else if ((DesiredClass ~= "MCoop2.MCoopMercenaryElite") || (InName ~= "mercenaryelite"))	{ SpawnClass = Class'MCoopMercenaryElite';	PawnType = 12; }
	else if ((DesiredClass ~= "MCoop2.MCoopNaliCow") || (InName ~= "nalicow"))			{ SpawnClass = Class'MCoopNaliCow';		PawnType = 13; }
	else if ((DesiredClass ~= "MCoop2.MCoopBabyCow") || (InName ~= "babycow"))			{ SpawnClass = Class'MCoopBabyCow';		PawnType = 13; }
	else if ((DesiredClass ~= "MCoop2.MCoopNaliRabbit") || (InName ~= "nalirabbit"))		{ SpawnClass = Class'MCoopNaliRabbit';		PawnType = 14; }
	else if ((DesiredClass ~= "MCoop2.MCoopPupae") || (InName ~= "pupae"))				{ SpawnClass = Class'MCoopPupae';		PawnType = 15; }
	else if ((DesiredClass ~= "MCoop2.MCoopQueen") || (InName ~= "queen"))				{ SpawnClass = Class'MCoopQueen';		PawnType = 16; }
	else if ((DesiredClass ~= "MCoop2.MCoopSkaarjTrooper") || (InName ~= "skaarjtrooper"))		{ SpawnClass = Class'MCoopSkaarjTrooper';	PawnType = 17; }
	else if ((DesiredClass ~= "MCoop2.MCoopSkaarjGunner") || (InName ~= "skaarjgunner"))		{ SpawnClass = Class'MCoopSkaarjGunner';	PawnType = 17; }
	else if ((DesiredClass ~= "MCoop2.MCoopSkaarjInfantry") || (InName ~= "skaarjinfantry"))	{ SpawnClass = Class'MCoopSkaarjInfantry';	PawnType = 17; }
	else if ((DesiredClass ~= "MCoop2.MCoopSkaarjOfficer") || (InName ~= "skaarjofficer"))		{ SpawnClass = Class'MCoopSkaarjOfficer';	PawnType = 17; }
	else if ((DesiredClass ~= "MCoop2.MCoopSkaarjSniper") || (InName ~= "skaarjsniper"))		{ SpawnClass = Class'MCoopSkaarjSniper';	PawnType = 17; }
	else if ((DesiredClass ~= "MCoop2.MCoopSkaarjWarrior") || (InName ~= "skaarjwarrior"))		{ SpawnClass = Class'MCoopSkaarjWarrior';	PawnType = 18; }
	else if ((DesiredClass ~= "MCoop2.MCoopSkaarjAssassin") || (InName ~= "skaarjassassin"))	{ SpawnClass = Class'MCoopSkaarjAssassin';	PawnType = 18; }
	else if ((DesiredClass ~= "MCoop2.MCoopSkaarjBerserker") || (InName ~= "skaarjberserker"))	{ SpawnClass = Class'MCoopSkaarjBerserker';	PawnType = 18; }
	else if ((DesiredClass ~= "MCoop2.MCoopIceSkaarj") || (InName ~= "iceskaarj"))			{ SpawnClass = Class'MCoopIceSkaarj';		PawnType = 18; }
	else if ((DesiredClass ~= "MCoop2.MCoopSkaarjLord") || (InName ~= "skaarjlord"))		{ SpawnClass = Class'MCoopSkaarjLord';		PawnType = 18; }
	else if ((DesiredClass ~= "MCoop2.MCoopSkaarjScout") || (InName ~= "skaarjscout"))		{ SpawnClass = Class'MCoopSkaarjScout';		PawnType = 18; }
	else if ((DesiredClass ~= "MCoop2.MCoopSlith") || (InName ~= "slith"))				{ SpawnClass = Class'MCoopSlith';		PawnType = 19; }
	else if ((DesiredClass ~= "MCoop2.MCoopSquid") || (InName ~= "squid"))				{ SpawnClass = Class'MCoopSquid';		PawnType = 20; }
	else if ((DesiredClass ~= "MCoop2.MCoopTentacle") || (InName ~= "tentacle"))			{ SpawnClass = Class'MCoopTentacle';		PawnType = 21; }
	else if ((DesiredClass ~= "MCoop2.MCoopTitan") || (InName ~= "titan"))				{ SpawnClass = Class'MCoopTitan';		PawnType = 22; }
	else if ((DesiredClass ~= "MCoop2.MCoopStoneTitan") || (InName ~= "stonetitan"))		{ SpawnClass = Class'MCoopStoneTitan';		PawnType = 22; }
	else if ((DesiredClass ~= "MCoop2.MCoopWarLord") || (InName ~= "warlord"))			{ SpawnClass = Class'MCoopWarLord';		PawnType = 23; }
	else if (DesiredClass ~= "UnrealShare.UnrealSpectator")
	{
		PawnType = Rand(24);
		switch ( PawnType )
		{
			case 0:
				SpawnClass = Class'MCoopBird';
				break;
			case 1:
				SpawnClass = Class'MCoopBiterFish';
				break;
			case 2:
				SpawnClass = Class'MCoopBlob';
				break;
			case 3:
				switch ( Rand(3) )
				{
					case 0:
						SpawnClass = Class'MCoopBrute';
						break;
					case 1:
						SpawnClass = Class'MCoopLesserBrute';
						break;
					case 2:
						SpawnClass = Class'MCoopBehemoth';
						break;
				}
				break;
			case 4:
				SpawnClass = Class'MCoopDevilFish';
				break;
			case 5:
				SpawnClass = Class'MCoopFly';
				break;
			case 6:
				SpawnClass = Class'MCoopGasBag';
				break;
			case 7:
				SpawnClass = Class'MCoopGiantGasBag';
				break;
			case 8:
				SpawnClass = Class'MCoopGiantManta';
				break;
			case 9:
				SpawnClass = Class'MCoopHorseFly';
				break;
			case 10:
				if (FRand() < 0.5)
					SpawnClass = Class'MCoopKrall';
				else
					SpawnClass = Class'MCoopKrallElite';
				break;
			case 11:
				if (FRand() < 0.5)
					SpawnClass = Class'MCoopManta';
				else
					SpawnClass = Class'MCoopCaveManta';
				break;
			case 12:
				if (FRand() < 0.5)
					SpawnClass = Class'MCoopMercenary';
				else
					SpawnClass = Class'MCoopMercenaryElite';
				break;
			case 13:
				if (FRand() < 0.5)
					SpawnClass = Class'MCoopNaliCow';
				else
					SpawnClass = Class'MCoopBabyCow';
				break;
			case 14:
				SpawnClass = Class'MCoopNaliRabbit';
				break;
			case 15:
				SpawnClass = Class'MCoopPupae';
				break;
			case 16:
				SpawnClass = Class'MCoopQueen';
				break;
			case 17:
				switch ( Rand(5) )
				{
					case 0:
						SpawnClass = Class'MCoopSkaarjTrooper';
						break;
					case 1:
						SpawnClass = Class'MCoopSkaarjGunner';
						break;
					case 2:
						SpawnClass = Class'MCoopSkaarjInfantry';
						break;
					case 3:
						SpawnClass = Class'MCoopSkaarjOfficer';
						break;
					case 4:
						SpawnClass = Class'MCoopSkaarjSniper';
						break;
				}
				break;
			case 18:
				switch ( Rand(6) )
				{
					case 0:
						SpawnClass = Class'MCoopSkaarjWarrior';
						break;
					case 1:
						SpawnClass = Class'MCoopSkaarjAssassin';
						break;
					case 2:
						SpawnClass = Class'MCoopSkaarjBerserker';
						break;
					case 3:
						SpawnClass = Class'MCoopIceSkaarj';
						break;
					case 4:
						SpawnClass = Class'MCoopSkaarjLord';
						break;
					case 5:
						SpawnClass = Class'MCoopSkaarjScout';
						break;
				}
				break;
			case 19:
				SpawnClass = Class'MCoopSlith';
				break;
			case 20:
				SpawnClass = Class'MCoopSquid';
				break;
			case 21:
				SpawnClass = Class'MCoopTentacle';
				break;
			case 22:
				if (FRand() < 0.5)
					SpawnClass = Class'MCoopTitan';
				else
					SpawnClass = Class'MCoopStoneTitan';
				break;
			case 23:
				SpawnClass = Class'MCoopWarlord';
				break;
			default:
				SpawnClass = Class'MCoopBird';
				break;
		}
	}
	else
	{
		InTeam = 0;
		PawnType = 255;

		if ( (DesiredClass ~= "MCoop2.MCoopNaliPriest") || (InName ~= "nalipriest") )
			SpawnClass = Class'MCoopNaliPriest';
		else if ( (DesiredClass ~= "MCoop2.MCoopNaliPlayer") || (DesiredClass ~= "UnrealI.naliplayer") || (InName ~= "naliplayer") )
			SpawnClass = Class'MCoopNaliPlayer';
		else if ( (DesiredClass ~= "MCoop2.MCoopSkaarjPlayer") || (DesiredClass ~= "UnrealI.SkaarjPlayer") || (InName ~= "skaarjplayer") )
			SpawnClass = Class'MCoopSkaarjPlayer';
		else if ( (DesiredClass ~= "MCoop2.MCoopFemaleOne") || (DesiredClass ~= "UnrealShare.FemaleOne") || (DesiredClass ~= "UPak.UPakFemaleOne") )
			SpawnClass = Class'MCoopFemaleOne';
		else if ( (DesiredClass ~= "MCoop2.MCoopFemaleTwo") || (DesiredClass ~= "UnrealI.FemaleTwo") || (DesiredClass ~= "UPak.UPakFemaleTwo") )
			SpawnClass = Class'MCoopFemaleTwo';
		else if ( (DesiredClass ~= "MCoop2.MCoopMaleOne") || (DesiredClass ~= "UnrealI.MaleOne") || (DesiredClass ~= "UPak.UPakMaleOne") )
			SpawnClass = Class'MCoopMaleOne';
		else if ( (DesiredClass ~= "MCoop2.MCoopMaleTwo") || (DesiredClass ~= "UnrealI.MaleTwo") || (DesiredClass ~= "UPak.UPakMaleTwo") )
			SpawnClass = Class'MCoopMaleTwo';
		else if ( (DesiredClass ~= "MCoop2.MCoopMaleThree") || (DesiredClass ~= "UnrealShare.MaleThree") || (DesiredClass ~= "UPak.UPakMaleThree") )
			SpawnClass = Class'MCoopMaleThree';
		else
			SpawnClass = MCoopGame.DefaultPlayerClass;
	}

	return SpawnClass;
}

//
// Log a player in.
// Fails login if you set the Error string.
// PreLogin is called before Login, but significant game time may pass before
// Login is called, especially if content is downloaded.
//
function PlayerPawn ServerLogin(string Portal, string Options, out string Error, class<PlayerPawn> SpawnClass)
{
	local Pawn PawnLink;
	local NavigationPoint StartSpot;
	local PlayerPawn NewPlayer, TestPlayer;
	local string PlayerIP, Port, InName, InClass, InSkin, InPassword, InFace;
	local byte InTeam, PawnType;
	local bool bPendingAdmin, bPendingAdminChild, bPendingAdminBaby;
	local LoginProtector LoginProtector;

	if ((Level.NetMode == NM_DedicatedServer) || (Level.NetMode == NM_ListenServer))
	{
		// get player's IP
		class'MCoopDLL'.static.GetIP( IPLoginOffset, PlayerIP, Port );

		// get the login options request
		foreach AllActors(class'LoginProtector', LoginProtector)
		{
			if ( LoginProtector.PlayerIP == PlayerIP )
			{
				InName	   = LoginProtector.LoginOptions.InName;
				InClass	   = LoginProtector.LoginOptions.InClass;
				SpawnClass = LoginProtector.LoginOptions.SpawnClass;
				InSkin	   = LoginProtector.LoginOptions.InSkin;
				InTeam	   = LoginProtector.LoginOptions.InTeam;
				InPassword = LoginProtector.LoginOptions.InPassword;
				PawnType   = LoginProtector.LoginOptions.PawnType;

				LoginProtector.bAllowLogin = False;
				LoginProtector.TimeLeft = LoginProtector.DelayTime;
				LoginProtector.TimeOutLeft = LoginProtector.DelayTime * 2;

				break;
			}
		}
	}

	// no LoginProtector actor for this player found...
	if ( InName == "" )
	{
		InName     = Left(MCoopGame.ParseOption(Options, "Name"), 20);
		InClass    = Left(MCoopGame.ParseOption(Options, "Class"), 30);
		InSkin	   = MCoopGame.ParseOption(Options, "Skin");
		InPassword = MCoopGame.ParseOption(Options, "Password");
		SpawnClass = ModifyLogin(InName, InClass, InTeam, PawnType);

		if (InName == "")	InName = MCoopGame.DefaultPlayerName;
	}

	InFace = "";

	ServerLog("Name: "$ InName);
	ServerLog("Desired Class: "$ InClass);
	ServerLog("Spawn Class: "$ string(SpawnClass));
	ServerLog("Team: "$ MCoopGame.Teams[InTeam].TeamName);
	ServerLog("Password: "$ InPassword);
	if ( !ClassIsChildOf(class'MCoopMonster', SpawnClass) )
		ServerLog("Skin: "$ InSkin);

	// Admin checks
	bPendingAdmin = (AdminPassword != "") && (InPassword == AdminPassword);
	bPendingAdminChild = (AdminChildPassword != "") && (InPassword == AdminChildPassword);
	bPendingAdminBaby = (AdminBabyPassword != "") && (InPassword == AdminBabyPassword);

	// Check if MaxPlayers limit is reached
	if ( (Level.NetMode != NM_Standalone) && (MCoopGame.MaxPlayers > 0) && (MCoopGame.NumPlayers >= MCoopGame.MaxPlayers) && !bPendingAdmin )
	{
		Error = MCoopGame.MaxedOutMessage;
		return None;
	}

	StartSpot = FindStart(SpawnClass, PawnType, InTeam, Portal);

	if (StartSpot == None)
	{
		Error = MCoopGame.FailedPlaceMessage;
		return None;
	}

	// Try to match up to existing unoccupied player in level, for savegames and coop level switching.
	if ( !ClassIsChildOf(class'MCoopMonster', SpawnClass) || (Level.NetMode == NM_Standalone) )
	{
		for (PawnLink=Level.PawnList; PawnLink!=None; PawnLink=PawnLink.nextPawn)
		{
			TestPlayer = PlayerPawn(PawnLink);
			if ((TestPlayer != None) && (TestPlayer.Player == None))
			{
				// Found matching unoccupied player, so use this one.
				NewPlayer = TestPlayer;
				break;
			}
		}
	}

	// If not found, spawn a new player.
	if (NewPlayer == None)
	{
		NewPlayer = Spawn(SpawnClass,,, StartSpot.Location, StartSpot.Rotation);
		if (NewPlayer != None)
			NewPlayer.ViewRotation = StartSpot.Rotation;
	}

	// Handle spawn failure.
	if (NewPlayer == None)
	{
		WarnLog("Couldn't find any point where to spawn player at");
		Error = MCoopGame.FailedSpawnMessage;
		return None;
	}

	if (InSkin != "")
		MCoopUnrealIPlayer(NewPlayer).static.SetMultiSkin(NewPlayer, InSkin, InFace, InTeam);

	if ( !MCoopGame.ChangeTeam(newPlayer, InTeam) )
	{
		Error = MCoopGame.FailedTeamMessage;
		return None;
	}

	// Init player's administrative privileges
	NewPlayer.Password = InPassword;
	NewPlayer.bAdmin = bPendingAdmin;
	MCoopUnrealIPlayer(NewPlayer).bAdminChild = bPendingAdminChild;
	MCoopUnrealIPlayer(NewPlayer).bAdminBaby = bPendingAdminBaby;
	if ( NewPlayer.bAdmin )
	{
		ServerLog("Administrator logged in");
		MCoopGame.BroadcastMessage(InName $" is administrator.");
	}
	else if ( MCoopUnrealIPlayer(NewPlayer).bAdminChild )
	{
		ServerLog("AdminChild logged in");
		MCoopGame.BroadcastMessage(InName $" is a moderator.");
	}
	else if ( MCoopUnrealIPlayer(NewPlayer).bAdminBaby )
	{
		ServerLog("AdminBaby logged in");
		MCoopGame.BroadcastMessage(InName $" is a summoner.");
	}

	// Init player's information.
	NewPlayer.ClientSetRotation(NewPlayer.Rotation);

	if ( !ChatLogger.AcceptName(InName) )
		InName = MCoopGame.DefaultPlayerName;
	NewPlayer.PlayerReplicationInfo.PlayerName = InName;

	// Init player's replication info.
	NewPlayer.GameReplicationInfo = MCoopGame.GameReplicationInfo;

	// If we are a server, broadcast a welcome message.
	if ( Level.NetMode==NM_DedicatedServer || Level.NetMode==NM_ListenServer )
	{
		MCoopGame.BroadcastMessage(NewPlayer.PlayerReplicationInfo.PlayerName $ MCoopGame.EnteredMessage, False);
		if ( Team(Self) != None )
			MCoopGame.BroadcastMessage(NewPlayer.PlayerReplicationInfo.PlayerName $" has joined "$ MCoopGame.Teams[InTeam].TeamName, False);
	}

	// Teleport-in effect.
	StartSpot.PlayTeleportEffect( NewPlayer, True );

	// Set the player's ID, IP and PawnType
	NewPlayer.PlayerReplicationInfo.PlayerID = MCoopGame.CurrentID;
	MCoopPlayerReplicationInfo(NewPlayer.PlayerReplicationInfo).PlayerID = MCoopGame.CurrentID++;
	MCoopPlayerReplicationInfo(NewPlayer.PlayerReplicationInfo).PlayerIP = PlayerIP;
	MCoopPlayerReplicationInfo(NewPlayer.PlayerReplicationInfo).PawnType = PawnType;

	// Log it.
	if ( MCoopGame.LocalLog != None )
		MCoopGame.LocalLog.LogPlayerConnect(NewPlayer);
	if ( MCoopGame.WorldLog != None )
		MCoopGame.WorldLog.LogPlayerConnect(NewPlayer);
	if ( ChatLogger.MCoopLogFile != None )
		ChatLogger.MCoopLogFile.LogPlayerConnect(NewPlayer);

	MCoopGame.NumPlayers++;

	return NewPlayer;
}

//
// Called after a successful login. This is the first place
// it is safe to call replicated functions on the PlayerPawn.
//
function ServerPostLogin(PlayerPawn NewPlayer)
{
	local int MaxClientRate;

	Log("-- Login complete: "$ NewPlayer.PlayerReplicationInfo.PlayerName $" -- IP Address: "$ MCoopPlayerReplicationInfo(NewPlayer.PlayerReplicationInfo).PlayerIP, 'MCoop');

	if ( IPLogger.bEnabled )
	{
		// Log player who's just entered the game
		IPLogger.LogIP(NewPlayer.PlayerReplicationInfo.PlayerName, MCoopPlayerReplicationInfo(NewPlayer.PlayerReplicationInfo).PlayerIP);
	}

	// Start player's music.
	NewPlayer.ClientSetMusic(Level.Song, Level.SongSection, Level.CdTrack, MTRAN_Fade);

	if ((Level.NetMode == NM_DedicatedServer) || (Level.NetMode == NM_ListenServer))
	{
		if ( AdvancedNetwork.bOptimizeMaxPlayers && IPLogger.IsALanPlayer(MCoopPlayerReplicationInfo(NewPlayer.PlayerReplicationInfo).PlayerIP) )
		{
			MCoopGame.MaxPlayers++;
			Log("-- LAN player found! MaxPlayers has been increased to... "$ MCoopGame.MaxPlayers $" --", 'MCoop');
		}
		else if ( AdvancedNetwork.bOptimizeConnection && (MCoopGame.NumPlayers > 0) )
		{
			MaxClientRate = AdvancedNetwork.ServerUploadSpeed / CountPlayers();

			if ( MaxClientRate > AdvancedNetwork.DefaultMaxClientRate )
				MaxClientRate = AdvancedNetwork.DefaultMaxClientRate;

			if ( MaxClientRate < AdvancedNetwork.DefaultMinClientRate )
				MaxClientRate = AdvancedNetwork.DefaultMinClientRate;

			if ( MaxClientRate != int(ConsoleCommand("get ipdrv.tcpnetdriver MaxClientRate")) )
			{
				Log("-- New MaxClientRate... "$ MaxClientRate $" --", 'MCoop');
				ConsoleCommand("set ipdrv.tcpnetdriver MaxClientRate "$ string(MaxClientRate));
				MCoopUnrealIPlayer(NewPlayer).InitialMaxClientRate = MaxClientRate;
				FixPlayersNetSpeed(MaxClientRate, NewPlayer);
			}
		}
	}
}

//
// Pawn exits.
//
function ServerLogout(Pawn Exiting)
{
	local int MaxClientRate;

	if ( Exiting.PlayerReplicationInfo == None )
		return;
	if ( Exiting.IsA('PlayerPawn') )
	{
		MCoopGame.NumPlayers--;
		ServerLog(Exiting.PlayerReplicationInfo.PlayerName $" left the game");
		ClearEnemy(Exiting);

		if ((Level.NetMode == NM_DedicatedServer) || (Level.NetMode == NM_ListenServer))
		{
			if ( AdvancedNetwork.bOptimizeMaxPlayers && IPLogger.IsALanPlayer(MCoopPlayerReplicationInfo(Exiting.PlayerReplicationInfo).PlayerIP) )
			{
				MCoopGame.MaxPlayers--;
				Log("-- LAN player found! MaxPlayers has been decreased to... "$ MCoopGame.MaxPlayers $" --", 'MCoop');
			}
			else if ( AdvancedNetwork.bOptimizeConnection && (MCoopGame.NumPlayers > 0) )
			{
				MaxClientRate = AdvancedNetwork.ServerUploadSpeed / CountPlayers();

				if ( MaxClientRate > AdvancedNetwork.DefaultMaxClientRate )
					MaxClientRate = AdvancedNetwork.DefaultMaxClientRate;

				if ( MaxClientRate < AdvancedNetwork.DefaultMinClientRate )
					MaxClientRate = AdvancedNetwork.DefaultMinClientRate;

				if ( MaxClientRate != int(ConsoleCommand("get ipdrv.tcpnetdriver MaxClientRate")) )
				{
					Log("-- New MaxClientRate... "$ MaxClientRate $" --", 'MCoop');
					ConsoleCommand("set ipdrv.tcpnetdriver MaxClientRate "$ string(MaxClientRate));
					FixPlayersNetSpeed(MaxClientRate, Exiting);
				}
			}
		}

		if ( ChatLogger.MCoopLogFile != None )
			ChatLogger.MCoopLogFile.LogPlayerDisconnect(Exiting);
		if ( MCoopUnrealIPlayer(Exiting).HUDInfo != None )
			MCoopUnrealIPlayer(Exiting).HUDInfo.Destroy();
		if ( MCoopUnrealIPlayer(Exiting).ScoreBoardInfo != None )
			MCoopUnrealIPlayer(Exiting).ScoreBoardInfo.Destroy();
		if ( MCoopUnrealIPlayer(Exiting).TeamScoreBoardInfo != None )
			MCoopUnrealIPlayer(Exiting).TeamScoreBoardInfo.Destroy();
	}
	if ( (Level.NetMode == NM_DedicatedServer) || (Level.NetMode == NM_ListenServer) )
		BroadcastMessage(Exiting.PlayerReplicationInfo.PlayerName$MCoopGame.LeftMessage, False);
	if ( MCoopGame.LocalLog != None )
		MCoopGame.LocalLog.LogPlayerDisconnect(Exiting);
	if ( MCoopGame.WorldLog != None )
		MCoopGame.WorldLog.LogPlayerDisconnect(Exiting);
	if ( MCoopGame.NumPlayers < 0 )
		MCoopGame.NumPlayers = 0;
}

// adjust NetSpeed of all the non-LAN players in the server
// eventually, SKIP SkipPlayer
function FixPlayersNetSpeed(int MaxClientRate, optional Pawn SkipPlayer)
{
	local Pawn aPawn;

	for ( aPawn=Level.PawnList; aPawn!=None; aPawn=aPawn.nextPawn )
	{
		if ( (MCoopUnrealIPlayer(aPawn) != None) && (NetConnection(PlayerPawn(aPawn).Player) != None) && (aPawn.PlayerReplicationInfo != None)
			&& !IPLogger.IsALanPlayer(MCoopPlayerReplicationInfo(aPawn.PlayerReplicationInfo).PlayerIP) )
		{
			if ( (SkipPlayer != None) && (aPawn == SkipPlayer) )
				continue;

			MCoopUnrealIPlayer(aPawn).AdjustNetSpeed(MaxClientRate);
		}
	}
}

// return the number of non-LAN players in the server
function int CountPlayers()
{
	local Pawn aPawn;
	local int result;

	for ( aPawn=Level.PawnList; aPawn!=None; aPawn=aPawn.nextPawn )
	{
		if ( (PlayerPawn(aPawn) != None) && (NetConnection(PlayerPawn(aPawn).Player) != None) && (aPawn.PlayerReplicationInfo != None)
			&& !IPLogger.IsALanPlayer(MCoopPlayerReplicationInfo(aPawn.PlayerReplicationInfo).PlayerIP) )
			result++;
	}

	return result;
}

function ClearEnemy(Pawn Exiting)
{
	local Pawn aPawn;

	for ( aPawn=Level.PawnList; aPawn!=None; aPawn=aPawn.nextPawn )
	{
		if ( aPawn.Enemy == Exiting )
		{
			aPawn.Enemy = None;
			if ( ScriptedPawn(aPawn) != None )
			{
				if (ScriptedPawn(aPawn).OldEnemy == Exiting)
					ScriptedPawn(aPawn).OldEnemy = None;
				ScriptedPawn(aPawn).WhatToDoNext('Waiting', 'TurnFromWall');
			}
		}
	}
}

//
// Add bot to game.
//
function bool ServerAddBot()
{
}

//-------------------------------- Inventory ----------------------------------------
//
// Examine the passed player's inventory, and accept or discard each item.
// AcceptInventory needs to gracefully handle the case of some inventory
// being accepted but other inventory not being accepted (such as the default
// weapon).  There are several things that can go wrong: A weapon's
// AmmoType not being accepted but the weapon being accepted -- the weapon
// should be killed off. Or the player's selected inventory item, active
// weapon, etc. not being accepted, leaving the player weaponless or leaving
// the HUD inventory rendering messed up (AcceptInventory should pick another
// applicable weapon/item as current).
//
function ServerAcceptInventory(Pawn PlayerPawn)
{
	// Default accept all inventory except default weapon (spawned explicitly)
	// Initialize the inventory.
	MCoopGame.AddDefaultInventory(PlayerPawn);

	ServerLog("All inventory from "$ PlayerPawn.PlayerReplicationInfo.PlayerName $" is accepted");
}

//
// Spawn any default inventory for the player.
//
function ServerAddDefaultInventory(Pawn PlayerPawn)
{
	local bool bSpawnCustomInv;
	local class<MCoopInventory> MCoopInvClass;
	local MCoopInventory MCoopInv;
	local class<Weapon> BaseWeapon;
	local Weapon newWeapon;

	PlayerPawn.JumpZ = PlayerPawn.Default.JumpZ * MCoopGame.PlayerJumpZScaling();

	if ( MCoopUnrealIPlayer(PlayerPawn) != None )
	{
		bSpawnCustomInv = (MCoopUnrealIPlayer(PlayerPawn).InventoryClass != "") && (PlayerOptions.bSetDefaultInventory || PlayerOptions.bSetDefaultHealth);

		// spawn player's custom Inventory class if needed
		if ( bSpawnCustomInv )
		{
			MCoopInvClass = class<MCoopInventory>(DynamicLoadObject(MCoopUnrealIPlayer(PlayerPawn).InventoryClass, class'Class'));
			if ( MCoopInvClass == None )
				WarnLog("GameMutator.ServerAddDefaultInventory couldn't load class '"$ MCoopUnrealIPlayer(PlayerPawn).InventoryClass $"'");
			else
				MCoopInv = Spawn(class<MCoopInventory>(DynamicLoadObject(MCoopUnrealIPlayer(PlayerPawn).InventoryClass, class'Class')), Self);
		}
	}

	// set player's custom health
	if ( PlayerOptions.bSetDefaultHealth && (MCoopInv != None) )
		MCoopInv.SetDefaultHealth(PlayerPawn(PlayerPawn));

	// give custom inventory to player
	if ( PlayerOptions.bSetDefaultInventory && (MCoopInv != None) )
	{
		MCoopInv.AddDefaultStuff(PlayerPawn(PlayerPawn));
		MCoopInv.Destroy();
	}
	else
	{
		if ( MCoopMonster(PlayerPawn) != None )
			BaseWeapon = MCoopMonster(PlayerPawn).SpecialWeapon;
		else
			BaseWeapon = MCoopGame.BaseMutator.MutatedDefaultWeapon();

		// Spawn default weapon.
		if ( (MCoopGame.DefaultWeapon == None) || (PlayerPawn.FindInventoryType(BaseWeapon) != None) )
			return;

		newWeapon = Spawn(BaseWeapon);
		if (newWeapon != None)
		{
			newWeapon.Instigator = PlayerPawn;
			newWeapon.BecomeItem();
			PlayerPawn.AddInventory(newWeapon);
			newWeapon.BringUp();
			newWeapon.GiveAmmo(PlayerPawn);
			newWeapon.SetSwitchPriority(PlayerPawn);
			newWeapon.WeaponSet(PlayerPawn);
		}
	}
}

//-------------------------------- Other ----------------------------------------
//
// Return the 'best' player start for this player to start from.
// Re-implement for each game type.
//
function NavigationPoint FindStart(class<Pawn> SpawnClass, byte PawnType, byte Team, optional string incomingName)
{
	// if we're handling a MCoopMonster, let's use FindMonsterStart...
	if ( PawnType != 255 )
		return FindMonsterStart(SpawnClass, PawnType, Team, incomingName);

	// ...otherwise, FindPlayerStart :)
	return MCoopGame.FindPlayerStart(Team, incomingName);
}

function NavigationPoint ServerFindPlayerStart(byte Team, optional string incomingName)
{
}

function NavigationPoint FindMonsterStart(class<Pawn> SpawnClass, byte PawnType, byte Team, optional string incomingName)
{
	local NavigationPoint N, Candidate[250], Best;
	local float Score[250], BestScore;
	local int i, num;
	local Actor LevelActor;
	local class<SpawnTester> SpawnTesterClass;
	local SpawnTester SpawnTester;
	local ScriptedPawn StupidPawn;

	num = 0;

	// choose candidates
	for ( N=Level.NavigationPointList; N!=None; N=N.nextNavigationPoint )
	{
		if ( N.IsA('PathNode') )
		{
			if (num < 250)
				Candidate[num] = N;
			num++;
		}
		else if ( N.IsA('PlayerStart') || (N.IsA('MonsterStart') && (MonsterStart(N).PawnType == PawnType)) )
		{
			if (num < 250)
				Candidate[num] = N;
			else if ( Rand(num) < 250 )
				Candidate[Rand(250)] = N;
			num++;
		}
	}

	if (num > 250)
		num = 250;
	else if (num == 0)
	{
		WarnLog("This map is not path-noded");
		return None;
	}

	// assess candidates
	for ( i=0; i<num; i++ )
		Score[i] = num * 1000 * FRand(); // randomize

	// find the right size of the SpawnTester actor
	if ( (SpawnClass.default.CollisionRadius < 20) && (SpawnClass.default.CollisionHeight < 20) )
		SpawnTesterClass = Class'XSmallSpawnTester';
	else if ( (SpawnClass.default.CollisionRadius < 40) && (SpawnClass.default.CollisionHeight < 40) )
		SpawnTesterClass = Class'SmallSpawnTester';
	else if ( (SpawnClass.default.CollisionRadius < 60) && (SpawnClass.default.CollisionHeight < 60) )
		SpawnTesterClass = Class'MedSpawnTester';
	else if ( (SpawnClass.default.CollisionRadius < 80) && (SpawnClass.default.CollisionHeight < 80) )
		SpawnTesterClass = Class'BigSpawnTester';
	else
		SpawnTesterClass = Class'XBigSpawnTester';

	for ( i=0; i<num; i++ )
	{
		foreach VisibleCollidingActors(class'Actor', LevelActor, SpawnClass.default.CollisionRadius * 2, Candidate[i].Location, True)
		{
			// ServerLog(SpawnClass.default.CollisionRadius * 2 $"   LevelActor found near Candidate["$i$"] ("$ Candidate[i] $)": "$ LevelActor);
			Score[i] -= 100000.0;
		}

		SpawnTester = Spawn(SpawnTesterClass,,, Candidate[i].Location, Candidate[i].Rotation);
		if ( SpawnTester != None )
			SpawnTester.Destroy();
		else
			Score[i] -= 500000.0;
	}

	BestScore = Score[0];
	Best = Candidate[0];
	for ( i=1; i<num; i++ )
	{
		// ServerLog("Candidate["$ i $"] = "$ Score[i]);
		if (Score[i] > BestScore)
		{
			BestScore = Score[i];
			Best = Candidate[i];
		}
	}

	// if there is a scripted pawn being still right near the spawn point, destroy it
	foreach VisibleCollidingActors(class'ScriptedPawn', StupidPawn, SpawnClass.default.CollisionRadius * 2, Best.Location, True)
		if ( (StupidPawn.Velocity.X * StupidPawn.Velocity.X + StupidPawn.Velocity.Y * StupidPawn.Velocity.Y) < 1000 )
			StupidPawn.Destroy();

	return Best;
}

//
// Restart a player.
//
function bool ServerRestartPlayer(Pawn aPlayer)
{
	local NavigationPoint startSpot;
	local bool foundStart;

	if ( MCoopGame.bRestartLevel && (Level.NetMode != NM_DedicatedServer) && (Level.NetMode != NM_ListenServer) )
		return True;

	if ( MCoopPlayerReplicationInfo(aPlayer.PlayerReplicationInfo) != None )
		startSpot = FindStart(aPlayer.Class, MCoopPlayerReplicationInfo(aPlayer.PlayerReplicationInfo).PawnType, aPlayer.PlayerReplicationInfo.Team);
	if ( startSpot == None )
		return False;

	foundStart = aPlayer.SetLocation(startSpot.Location);
	if ( foundStart )
	{
		startSpot.PlayTeleportEffect(aPlayer, True);
		aPlayer.SetRotation(startSpot.Rotation);
		aPlayer.ViewRotation = aPlayer.Rotation;
		aPlayer.Acceleration = vect(0,0,0);
		aPlayer.Velocity = vect(0,0,0);
		aPlayer.Health = aPlayer.Default.Health;
		aPlayer.SetCollision(True, True, True);
		aPlayer.ClientSetLocation( startSpot.Location, startSpot.Rotation );
		aPlayer.bHidden = False;
		aPlayer.DamageScaling = aPlayer.Default.DamageScaling;
		aPlayer.SoundDampening = aPlayer.Default.SoundDampening;
		MCoopGame.AddDefaultInventory(aPlayer);
	}

	return foundStart;
}

function ServerKilled(Pawn killer, Pawn Other, Name damageType)
{
	local string killed, message, KillerWeaponName, VictimWeaponName, DeathMessage;

	if ( Other == None )
		return;

	// prevent 'Enemy = None' freeze
	ClearEnemy(Other);

	if ( Other.IsA('ScriptedPawn') && Other.bIsPlayer )
		Other.bIsPlayer = False;

	if ( Other.bIsPlayer )
	{
		if ( (Killer == None) || (Killer == Other) )
		{
			// suicide
			if ( MCoopGame.LocalLog != None )
				MCoopGame.LocalLog.LogSuicide(Other, damageType);
			if ( MCoopGame.WorldLog != None )
				MCoopGame.WorldLog.LogSuicide(Other, damageType);
			message = MCoopGame.KillMessage(damageType, Other);
			MCoopGame.BroadcastMessage(Other.PlayerReplicationInfo.PlayerName $ message, False, 'DeathMessage');
		}
		else
		{
			if ( Killer.Weapon != None )
			{
				KillerWeaponName = Killer.Weapon.ItemName;
				DeathMessage = Killer.Weapon.DeathMessage;
			}
			if ( Other.Weapon != None )
				VictimWeaponName = Other.Weapon.ItemName;

			if ( MCoopGame.LocalLog != None )
				MCoopGame.LocalLog.LogKill(Killer.PlayerReplicationInfo.PlayerID, Other.PlayerReplicationInfo.PlayerID, KillerWeaponName, VictimWeaponName, damageType);
			if ( MCoopGame.WorldLog != None )
				MCoopGame.WorldLog.LogKill(Killer.PlayerReplicationInfo.PlayerID, Other.PlayerReplicationInfo.PlayerID, KillerWeaponName, VictimWeaponName, damageType);

			killed = Other.PlayerReplicationInfo.PlayerName;
			if ( (DamageType == 'SpecialDamage') && (MCoopGame.SpecialDamageString != "") )
			{
				message = MCoopGame.ParseKillMessage(Killer.PlayerReplicationInfo.PlayerName, Other.PlayerReplicationInfo.PlayerName, KillerWeaponName, MCoopGame.SpecialDamageString);
				MCoopGame.BroadcastMessage(message, False, 'DeathMessage');
			}
			else if ( MCoopGame.bClassicDeathmessages || (Killer.Weapon == None) || ((DamageType != Killer.Weapon.MyDamageType) && (DamageType != Killer.Weapon.AltDamageType)) )
			{
				message = killer.KillMessage(damageType, Other);
				MCoopGame.BroadcastMessage(killed$message, False, 'DeathMessage');
			}
			else
			{
				message = MCoopGame.ParseKillMessage(Killer.PlayerReplicationInfo.PlayerName, Other.PlayerReplicationInfo.PlayerName, KillerWeaponName,	DeathMessage);
				MCoopGame.BroadcastMessage(message, False, 'DeathMessage');
			}
		}
	}

	MCoopGame.ScoreKill(killer, Other);
}

function ServerScoreKill(Pawn Killer, Pawn Other)
{
	Other.DieCount++;
	if ( (Other.PlayerReplicationInfo != None) && (MCoopPlayerReplicationInfo(Other.PlayerReplicationInfo) != None) )
		MCoopPlayerReplicationInfo(Other.PlayerReplicationInfo).Deaths += 1.0;

	if ( (killer == None) || (killer == Other) )
		Other.PlayerReplicationInfo.Score -= 1.0;
	else if ( killer != None )
	{
		killer.killCount++;
		if ( killer.PlayerReplicationInfo != None )
			killer.PlayerReplicationInfo.Score += 1.0;
	}
}

function bool ServerCanSpectate(Pawn Viewer, Actor ViewTarget)
{
	return True;
}

//
// Use reduce damage for teamplay modifications, etc.
//
function int ServerReduceDamage(int Damage, Name DamageType, Pawn injured, Pawn instigatedBy)
{
}

//
// Return whether an item should respawn.
//
function bool ServerShouldRespawn(Actor Other)
{
	if ( Other.IsA('Weapon') )
	{
		if ( !Inventory(Other).bHeldItem && (Inventory(Other).ReSpawnTime != 0) )
		{
			if ( RespawnOptions.bRespawnWeapon )
				Inventory(Other).RespawnTime = RespawnOptions.RespawnTimeWeapon;
			else
				Inventory(Other).RespawnTime = Inventory(Other).default.RespawnTime;
			return True;
		}
		return False;
	}
	else if ( Other.IsA('Ammo') )
	{
		if ( !Inventory(Other).bHeldItem && (Inventory(Other).RespawnTime != 0) )
		{
			if ( RespawnOptions.bRespawnAmmo )
				Inventory(Other).RespawnTime = RespawnOptions.RespawnTimeAmmo;
			else
				Inventory(Other).RespawnTime = Inventory(Other).default.RespawnTime;
			return True;
		}
		return False;
	}
	else if ( Other.IsA('Health') )
	{
		if ( !Inventory(Other).bHeldItem && (Inventory(Other).RespawnTime != 0) )
		{
			if ( RespawnOptions.bRespawnHealth )
				Inventory(Other).RespawnTime = RespawnOptions.RespawnTimeHealth;
			else
				Inventory(Other).RespawnTime = Inventory(Other).default.RespawnTime;
			return True;
		}
		return False;
	}
	else if ( Other.IsA('Inventory') )
	{
		if ( !Inventory(Other).bHeldItem && (Inventory(Other).RespawnTime != 0) )
		{
			if ( RespawnOptions.bRespawnInventory )
				Inventory(Other).RespawnTime = RespawnOptions.RespawnTimeInventory;
			else
				Inventory(Other).RespawnTime = Inventory(Other).default.RespawnTime;
			return True;
		}
		return False;
	}
}

//
// Called when pawn has a chance to pick Item up (i.e. when
// the pawn touches a weapon pickup). Should return true if
// he wants to pick it up, false if he does not want it.
//
function bool ServerPickupQuery(Pawn Other, Inventory Item)
{
	if ( !bAlwaysPickUp && Other.IsA('MCoopMonster') && !Other.IsA('MCoopSkaarjTrooper')
		&& (Item.IsA('Weapon') || Item.IsA('Ammo')) )
	{
		return False;
	}
	else
	{
		if ( Other.Inventory == None )
			return True;
		else
			return !Other.Inventory.HandlePickupQuery(Item);
	}
}

// Return the player jumpZ scaling for this gametype
function float ServerPlayerJumpZScaling()
{
	return 1.0;
}

//
// Try to change a player's name.
//
function ServerChangeName(Pawn Other, coerce string NewName, bool bNameChange)
{
	if ( Other.bIsPlayer )
	{
		if ( PlayerOptions.bDenyNameChange )
		{
			Other.ClientMessage("Changing nicknames is not allowed in this game :P");
			return;
		}

		if ( !ChatLogger.AcceptName(NewName) )
		{
			Other.ClientMessage("Your name sucks... choose another one :P");
			return;
		}
	}

	NewName = class'MCoopDLL'.static.Trim(NewName);
	if ( bNameChange )
	{
		MCoopGame.BroadcastMessage(Other.PlayerReplicationInfo.PlayerName $ MCoopGame.GlobalNameChange $ NewName, False);
		ServerLog(Other.PlayerReplicationInfo.PlayerName $" changed name to "$ NewName);

		if ( ChatLogger.MCoopLogFile != None )
			ChatLogger.MCoopLogFile.LogEventString(ChatLogger.MCoopLogFile.GetShortTime()$Chr(9)$ Other.PlayerReplicationInfo.PlayerName $" changed name to "$ NewName);
	}

	Other.PlayerReplicationInfo.PlayerName = NewName;
}

//
// Return whether a team change is allowed.
//
function bool ServerChangeTeam(Pawn Other, int N)
{
	if ( (Other.PlayerReplicationInfo.Team != 255) && MCoopGame.bNoTeamChanges )
		return False;

	Other.PlayerReplicationInfo.Team = N;
	return True;
}

//
// Play an inventory respawn effect.
//
function float ServerPlaySpawnEffect(Inventory Inv)
{
	Spawn(Class'ReSpawn',,, Inv.Location);
	return 0.3;
}

//
// Send a player to a URL.
//
function ServerSendPlayer(PlayerPawn aPlayer, string URL)
{
	aPlayer.ClientTravel(URL, TRAVEL_Relative, True);
}

//
// Play a teleporting special effect.
//
function ServerPlayTeleportEffect(Actor Incoming, bool bOut, bool bSound)
{
	local PawnTeleportEffect PTE;

	if ( Incoming.IsA('Pawn') )
	{
		if ( bSound )
		{
			PTE = Spawn(class'PawnTeleportEffect',,, Incoming.Location, Incoming.Rotation);
			PTE.Initialize(Pawn(Incoming), bOut);
			if ( Incoming.IsA('PlayerPawn') )
				PlayerPawn(Incoming).SetFOVAngle(170);
			Incoming.PlaySound(sound'Teleport1',, 10);
		}
	}
}

//
// Restart the game.
//
function ServerRestartGame()
{
	local MCoopMapList myList;
	local string NextMap;

	// multipurpose don't restart variable
	if ( MCoopGame.bDontRestart )
		return;

	ServerLog("Restart Game");

	// close the chatlog file
	if ( ChatLogger.MCoopLogFile != None )
	{
		ChatLogger.MCoopLogFile.StopLog();
		ChatLogger.MCoopLogFile.Destroy();
		ChatLogger.MCoopLogFile = None;
	}

	// these server travels should all be relative to the current URL
	if ( MCoopGame.bChangeLevels && !MCoopGame.bAlreadyChanged && (MapListType != None) )
	{
		// open a the nextmap actor for this game type and get the next map
		MCoopGame.bAlreadyChanged = True;
		myList = Spawn(MapListType);
		NextMap = myList.GetNextMap();
		myList.Destroy();
		if ( NextMap == "" )
			NextMap = MCoopGame.GetMapName(MCoopGame.MapPrefix, NextMap, 1);
		if ( NextMap != "" )
		{
			ServerLog("Changing to "$ NextMap);
			Level.ServerTravel(NextMap, False);
			return;
		}
	}

	Level.ServerTravel("?Restart", False);
}

//
// End of game.
//
function ServerEndGame(string Reason)
{
	local Actor A;
	local int WinnerTeam, HighestScore, LeastDeaths, CountTeam;
	local TeamInfo TI;
	local PlayerReplicationInfo PRI;
	local Actor Winner;
	local Pawn aPawn;

	// don't end game if not really ready
	if ( !MCoopGame.SetEndCams(Reason) )
	{
		MCoopGame.bOverTime = True;
		return;
	}

	MCoopGame.bGameEnded = True;
	MCoopGame.RemainingTime = -1;	// use timer to force restart

	foreach AllActors(class'Actor', A, 'EndGame')
		A.Trigger(Self, None);

	// close the chatlog file
	if (ChatLogger.MCoopLogFile != None)
	{
		ChatLogger.MCoopLogFile.StopLog();
		ChatLogger.MCoopLogFile.Destroy();
		ChatLogger.MCoopLogFile = None;
	}
	if (MCoopGame.LocalLog != None)
	{
		MCoopGame.LocalLog.LogGameEnd(Reason);
		MCoopGame.LocalLog.StopLog();
		MCoopGame.LocalLog.Destroy();
		MCoopGame.LocalLog = None;
	}
	if (MCoopGame.WorldLog != None)
	{
		MCoopGame.WorldLog.LogGameEnd(Reason);
		MCoopGame.WorldLog.StopLog();
		MCoopGame.WorldLog.Destroy();
		MCoopGame.WorldLog = None;
	}

	// What's the winning team? (code inspired by UTeamFix)
	if ( MCoopGame.bTeamGame )
	{
		HighestScore = -65000;
		WinnerTeam = -1;
		CountTeam = -1;
		foreach AllActors(class'TeamInfo', TI)
		{
			CountTeam++;
			if (TI.Score > HighestScore)
			{
				HighestScore = TI.Score;
				WinnerTeam = CountTeam;
			}
		}
	}

	if ( !MCoopGame.bTeamGame || (WinnerTeam != -1) )
	{
		// Who is the winner? (code inspired by UTeamFix)
		HighestScore = -65000;
		LeastDeaths  = 65000;

		foreach AllActors(class'PlayerReplicationInfo', PRI)
		{
			if ( !PRI.bIsSpectator && !MCoopGame.bTeamGame || (PRI.Team == WinnerTeam) )
				if ( (MCoopPlayerReplicationInfo(PRI) != None) && ( (PRI.Score > HighestScore)
					|| ((PRI.Score == HighestScore)	&& (MCoopPlayerReplicationInfo(PRI).Deaths < LeastDeaths)) ) )
				{
					HighestScore = PRI.Score;
					LeastDeaths = MCoopPlayerReplicationInfo(PRI).Deaths;
					Winner = PRI.Owner;
				}
		}
	}

	if ( Winner != None )
	{
		for ( aPawn=Level.PawnList; aPawn!=None; aPawn=aPawn.nextPawn )
		{
			if ( PlayerPawn(aPawn) != None )
			{
				PlayerPawn(aPawn).ViewTarget = Winner;
				PlayerPawn(aPawn).bBehindView = True;
				if ( MCoopGame.bTeamGame )
					aPawn.ClientMessage(Pawn(Winner).PlayerReplicationInfo.TeamName $" won the game!!", 'LowCriticalEvent', True);
				else
					aPawn.ClientMessage(Pawn(Winner).PlayerReplicationInfo.PlayerName $" won the game!!", 'LowCriticalEvent', True);
			}
		}
	}
}


//------------------------------ Exec Commands ------------------------------------
function SetAdminPw(string NewPassword)
{
	if ( Level.Netmode == NM_Standalone )
		return;
	if ( NewPassword != "" )
	{
		AdminPassword = NewPassword;
		SaveConfig();
		ServerLog("AdminPassword changed to "$ NewPassword);
	}
}

function SetAdminChildPw(string NewPassword)
{
	if ( Level.Netmode == NM_Standalone )
		return;
	if ( NewPassword != "" )
	{
		AdminChildPassword = NewPassword;
		SaveConfig();
		ServerLog("AdminChildPassword changed to "$ NewPassword);
	}
}

function SetAdminBabyPw(string NewPassword)
{
	if ( Level.Netmode == NM_Standalone )
		return;
	if ( NewPassword != "" )
	{
		AdminBabyPassword = NewPassword;
		SaveConfig();
		ServerLog("AdminBabyPassword changed to "$ NewPassword);
	}
}

function SetGamePw(string NewPassword)
{
	local int Offset;

	if ( Level.Netmode == NM_Standalone )
		return;
	if ( NewPassword != "" )
	{
		GamePassword = NewPassword;
		SaveConfig();
		if ( InStr(Caps(MCoopGame.GameReplicationInfo.ServerName), "PWD") == -1 )
		{
			if ( Right(Caps(MCoopGame.GameReplicationInfo.ServerName), 7) == "[MCOOP]" )
				MCoopGame.GameReplicationInfo.ServerName = MCoopGame.GameReplicationInfo.ServerName $"[Pwd]";
			else
				MCoopGame.GameReplicationInfo.ServerName = MCoopGame.GameReplicationInfo.ServerName $" [Pwd]";
		}
		ServerLog("GamePassword changed to "$ NewPassword);
	}
	else
	{
		GamePassword = "";
		Offset = InStr(Caps(MCoopGame.GameReplicationInfo.ServerName), "PWD");
		if ( Offset != -1 )
			MCoopGame.GameReplicationInfo.ServerName = Left(MCoopGame.GameReplicationInfo.ServerName, Offset);
		SaveConfig();
		ServerLog("GamePassword deleted");
	}
}

defaultproperties
{
				AdvancedNetwork=(ServerUploadSpeed=16000,DefaultMinClientRate=2600,DefaultMaxClientRate=10000)
				MaxFailedAdminLogins=5
				AdminName="Admin"
				BanError="You have been banned from this server... keep out :P"
				TempBanError="Okay... now you'll be glad to know that you've been auto-banned until map change :P"
}
